static const char rcsid[] = "$Id: log.c,v 1.11 2000/09/30 19:52:35 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
/* Code from this file sometimes works with root privileges */

#include "includes.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "log.h"
#include "response.h"

#ifdef SPIPV6
# include "spipv6.h"
#endif

struct log_struct {
	char *name;
	int num;
};

const struct log_struct log_facilities[] = {
	{"daemon", LOG_DAEMON},
	{"local0", LOG_LOCAL0},
	{"local1", LOG_LOCAL1},
	{"local2", LOG_LOCAL2},
	{"local3", LOG_LOCAL3},
	{"local4", LOG_LOCAL4},
	{"local5", LOG_LOCAL5},
	{"local6", LOG_LOCAL6},
	{"local7", LOG_LOCAL7},
	{"mail", LOG_MAIL},
	{"user", LOG_USER},
	{NULL, 0}
};

const struct log_struct log_priorities[] = {
	{"emerg", LOG_EMERG},
	{"alert", LOG_ALERT},
	{"crit", LOG_CRIT},
	{"err", LOG_ERR},
	{"warning", LOG_WARNING},
	{"notice", LOG_NOTICE},
	{"info", LOG_INFO},
	{"debug", LOG_DEBUG},
	{NULL, 0}
};


int pop_debug = 0;
int pop_facility = POP_FACILITY;
int pop_priority = POP_PRIORITY;
char logpriority[64];

extern char username[];
extern char ahname[];
#ifdef MAPPING
extern char mapusername[];
extern int domapping;
#endif

#ifdef LOG_EXTEND
int authlog_fd = -1;
#endif

void pop_openlog(void) { /* This function works with root privileges */
	openlog(POP_IDENT, LOG_PID, pop_facility);
}

void pop_closelog() { /* This function works with root privileges */
	closelog();
}

void pop_log(int priority, const char *format, ...) {
	/* This function works with root privileges */
	char tmp[1536];
	va_list lst;

	va_start(lst, format);
	vsnprintf(tmp, sizeof(tmp), format, lst);
	va_end(lst);
	syslog(priority, "%s", tmp);
}

void pop_log_dbg(int priority, const char *format, ...) {
	char tmp[1536];
	va_list lst;

	if (!pop_debug)
		return;
	va_start(lst, format);
	vsnprintf(tmp, sizeof(tmp), format, lst);
	va_end(lst);
	syslog(priority, "%s", tmp);
}


void pop_error(const char *name) {
	pop_log(pop_priority, "%.200s: %.200s", name, strerror(errno));
}

void pop_error_dbg(const char *name) {
	if (pop_debug)
		pop_error(name);
}

int search_log_struct(char *name, const struct log_struct table[]) {
	/* This function works with root privileges */
	int i = 0;
	
	while ((table[i].name != NULL)) {
		if (strcasecmp(table[i].name, name) == 0)
			break;
		i++;
	}
	return (table[i].name == NULL) ? -1 : table[i].num;
}

int check_logpriority(void *priority) {
	/* This function works with root privileges */
	char *tmp;
	int tmp_facility = pop_facility, tmp_priority = pop_priority;
	
	if ((tmp = strchr((char *) priority, '.')) != NULL)
		*tmp = 0;
	if ((tmp_facility = search_log_struct((char *) priority,
					      log_facilities)) < 0)
		return -1;
	if (tmp) {
		if ((tmp_priority = search_log_struct(tmp + 1,
						      log_priorities)) < 0)
			return -1;
		*tmp = '.';
	}
	pop_priority = tmp_priority;
	if (pop_facility != tmp_facility) {
		pop_facility = tmp_facility;
		pop_closelog();
		pop_openlog();
	}
	return 0;
}

#ifdef LOG_EXTEND

void authlog_openlog(char* path) {
/* This function works with root privileges */
	if ((authlog_fd = open(path, O_WRONLY | O_APPEND)) < 0)
		authlog_fd = open(path, O_WRONLY | O_APPEND | O_CREAT | O_EXCL,
			0600);
	if (authlog_fd < 0) {
		pop_log(pop_priority, "authlog: can't open or create: %.1024s",
			path);
		pop_error("authlog: open");
	}
}

void authlog_closelog(void) { /* This function works with root privileges */
	if (authlog_fd >= 0)
		close(authlog_fd);
	authlog_fd = -1;
}

void authlog_log(void) { /* This function works with root privileges */
	char tmp[1536];
	char vhost_name[384];
	
# ifdef SPIPV6
	union sp_sockaddr socket_addr;
	char ntopbuff[100];
# else
	struct sockaddr_in socket_addr;
# endif
	socklen_t socket_addr_len;

	if (authlog_fd < 0)
		return;
	socket_addr_len = sizeof(socket_addr);
	if (getsockname(0, (struct sockaddr *) &socket_addr,
		&socket_addr_len) < 0) {
		pop_error("getsockname");
		send_error("fatal error");
		exit(1);
	}
# ifdef SPIPV6
	if (socket_addr.saddr_in6.sin6_family == AF_INET6)
		snprintf(vhost_name, sizeof(vhost_name), "%.100s",
			(strncmp(inet_ntop(AF_INET6, 
				&socket_addr.saddr_in6.sin6_addr, ntopbuff,
				sizeof(ntopbuff)), "::ffff:", 7) == 0) ? 
			ntopbuff + 7 : ntopbuff);
	else
		snprintf(vhost_name, sizeof(vhost_name), "%.100s",
		    inet_ntoa(socket_addr.saddr_in.sin_addr));
# else
	snprintf(vhost_name, sizeof(vhost_name), "%.100s",
	    inet_ntoa(socket_addr.sin_addr));
# endif /* SPIPV6 */
# ifdef MAPPING
	if ((domapping) && (mapusername[0]))
		snprintf(tmp, sizeof(tmp), "User:%.40s Mapped:%.40s "
			"Remote:%.40s VHost:%.384s\n", username, mapusername,
			ahname, vhost_name);
	else
		snprintf(tmp, sizeof(tmp), "User:%.40s Remote:%.40s "
			"VHost:%.384s\n", username, ahname, vhost_name);
# else
	snprintf(tmp, sizeof(tmp), "User:%.40s Remote:%.40s VHost:%.384s\n",
		username, ahname, vhost_name);
# endif
	write(authlog_fd, tmp, strlen(tmp));
	authlog_closelog();
}
#endif /* LOG_EXTEND */

/* ahname is a 384-byte buffer */
#if defined(LOG_EXTEND) || defined(LOG_CONNECT) || defined(STANDALONE)
# ifdef RESOLVE_HOSTNAME
#  ifdef SPIPV6
void make_ahname(union sp_sockaddr *peeraddr, char *ahname) {
 /* This function works with root privileges */
	struct hostent *hentname;
	char ntopbuff[100];

	if (peeraddr->saddr_in6.sin6_family == AF_INET6) {
		if ((hentname = gethostbyaddr((char *)&peeraddr->saddr_in6.\
			sin6_addr, sizeof(union sp_sockaddr), AF_INET6)) != 
			NULL)
			snprintf(ahname, 384, "%.256s (%.100s)",
				hentname->h_name, (strncmp(inet_ntop(AF_INET6,
				&peeraddr->saddr_in6.sin6_addr, ntopbuff,
				sizeof(ntopbuff)), "::ffff:", 7) == 0) ? 
				ntopbuff + 7 : ntopbuff);
		else
			snprintf(ahname, 384, "%.100s",
				(strncmp(inet_ntop(AF_INET6,
				&peeraddr->saddr_in6.sin6_addr, ntopbuff,
				sizeof(ntopbuff)), "::ffff:", 7) == 0) ?
				ntopbuff + 7 : ntopbuff);
	} else {
		if ((hentname = gethostbyaddr((char *)&peeraddr->saddr_in.\
			sin_addr.s_addr, sizeof(peeraddr->saddr_in.sin_addr.\
			s_addr), AF_INET)) != NULL)
			snprintf(ahname, 384, "%.256s (%.100s)",
				hentname->h_name,
				inet_ntoa(peeraddr->saddr_in.sin_addr));
		else
			snprintf(ahname, 384, "%.100s",
			    inet_ntoa(peeraddr->saddr_in.sin_addr));

	}
#  else
void make_ahname(struct sockaddr_in *peeraddr, char *ahname) {
	struct hostent *hentname;

	if ((hentname = gethostbyaddr((char *)&peeraddr->sin_addr.s_addr,
		sizeof(peeraddr->sin_addr.s_addr), AF_INET)) != NULL)
		snprintf(ahname, 384, "%.256s (%.100s)", hentname->h_name,
		    inet_ntoa(peeraddr->sin_addr));
	else
		snprintf(ahname, 384, "%.100s",
		    inet_ntoa(peeraddr->sin_addr));
#  endif
# else /* RESOLVE_HOSTNAME */
#  ifdef SPIPV6
void make_ahname(union sp_sockaddr *peeraddr, char *ahname) {
	char ntopbuff[100];

	if (peeraddr->saddr_in6.sin6_family == AF_INET6)
		snprintf(ahname, 384, "%.100s", (strncmp(inet_ntop(AF_INET6,
			&peeraddr->saddr_in6.sin6_addr, ntopbuff,
			sizeof(ntopbuff)), "::ffff:", 7) == 0) ?
			ntopbuff + 7 : ntopbuff);
	else
		snprintf(ahname, 384, "%.100s",
		    inet_ntoa(peeraddr->saddr_in.sin_addr));
#  else
void make_ahname(struct sockaddr_in *peeraddr, char *ahname) {
	snprintf(ahname, 384, "%.100s",
	    inet_ntoa(peeraddr->sin_addr));
#  endif
# endif /* RESOLVE_HOSTNAME */
}
#endif /* LOG_EXTEND || LOG_CONNECT || STANDALONE */

void no_such_user(void) {

#ifdef MAPPING
	if (domapping)
# ifdef LOG_EXTEND
		pop_log(pop_priority, "authentication failed: "
			"no such user: %.40s - %.384s", mapusername, ahname);
# else
		pop_log(pop_priority, "authentication failed: "
			"no such user: %.40s", mapusername);
# endif
	else
# ifdef LOG_EXTEND
		pop_log(pop_priority, "authentication failed: "
			"no such user: %.40s - %.384s", username, ahname);
# else
		pop_log(pop_priority, "authentication failed: "
			"no such user: %.40s", username);				
# endif
#else /* MAPPING */
# ifdef LOG_EXTEND
	pop_log(pop_priority, "authentication failed: "
		"no such user: %.40s - %.384s", username, ahname);
# else
	pop_log(pop_priority, "authentication failed: "
		"no such user: %.40s", username);				
# endif
#endif /* MAPPING */
}
