/* $Id: mydbm.c,v 1.3 2000/08/23 11:05:19 jurekb Exp $ */
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef MYSBM_SPDBM
# include "includes.h"
# include "mydbm.h"
# include "log.h"
#endif

#include <sys/stat.h>
#include <fcntl.h>

#ifdef GDBM
# include <gdbm.h>
#else /* GDBM */
# ifdef HAVE_NDBM_H
#  include <ndbm.h>
# else /* HAVE_NDBM_H */
#  ifdef HAVE_DBM_H
#   include <dbm.h>
#  endif /* HAVE_DBM_H */
# endif /* HAVE_NDBM_H */
#endif /* GDBM */

#ifdef GDBM
int db_mode[] = {
	GDBM_READER, /* reader mode - DB_READER */
	GDBM_WRITER, /* writer mode - DB_WRITER */
	GDBM_WRCREAT, /* create database if it doesn't exist - DB_CREAT */
	GDBM_NEWDB /* create database regardless if one exists - DB_NEWDB */
};
#else
int db_mode[] = {
	O_RDONLY, 
	O_RDWR,
	O_RDWR | O_CREAT,
	O_RDWR | O_CREAT | O_TRUNC
};
#endif

int mydbm_open(MYDBM *db, char *filename, int mode) {
	int flags;

	if ((mode > 3) || (mode < 0)) {
#ifdef MYDBM_SPDBM
		fprintf(stderr, "mydbm: unexpected error - contact author "
			"of program\n");
#else
		pop_log(pop_priority, "mydbm: unexpected error - contact "
			"author of program");
#endif
		return -1;
	}
	flags = db_mode[mode];
#ifdef GDBM
	if (((*db) = gdbm_open(filename, 1024, flags, (mode >= 2) ? 0600 : 0,
		NULL)) == NULL) {
# ifdef MYDBM_SPDBM
		fprintf(stderr, "mydbm: can't open/create GDBM database: %s\n",
			filename);
		fprintf(stderr, "mydbm: %s\n", gdbm_strerror(gdbm_errno));
# else
		pop_log(pop_priority, "mydbm: can't open/create GDBM "
			"database: %.1024s", filename);
		pop_log(pop_priority, "mydbm: %.512s",
			gdbm_strerror(gdbm_errno));
# endif
	}
#else /* GDBM */
	if (((*db) = dbm_open(filename, flags, (mode >= 2) ? 0600 : 0)) == NULL)
# ifdef MYDBM_SPDBM
		fprintf(stderr, "mydbm: can't open/create database: %s\n",
			filename);
# else
		pop_log(pop_priority, "mydbm: can't open/create database: "
			"%.1024s", filename);
# endif
#endif /* GDBM */
	return (((*db) == NULL) ? -1 : 0);
}

void mydbm_close(MYDBM *db) {
#ifdef GDBM
	gdbm_close(*db);
#else
	dbm_close(*db);
#endif
}

int mydbm_fetch(MYDBM *db, char *key, char *value, int *valuesize) {
	datum akey, avalue;
	
	akey.dsize = strlen(akey.dptr = key);
#ifdef GDBM
	avalue = gdbm_fetch((*db), akey);
#else
	avalue = dbm_fetch((*db), akey);
#endif
	if (avalue.dptr == NULL)
		return -1;
	if (avalue.dsize >= (*valuesize)) {
#ifdef GDBM
		free(avalue.dptr);
#endif
#ifdef MYDBM_SPDBM
		fprintf(stderr, "mydbm: fatal: value is too long - damaged "
			"database\n");
#else
		pop_log(pop_priority, "mydbm: fatal: value is too long - "
			"damaged database");
#endif
		return -2;
	}
	memcpy(value, avalue.dptr, avalue.dsize);
	memset(avalue.dptr, 0, avalue.dsize);
#ifdef GDBM
	free(avalue.dptr);
#endif
	value[avalue.dsize] = 0;
	(*valuesize) = avalue.dsize;
	return 0;
}

int mydbm_exists(MYDBM *db, char *key) {
#ifndef GDBM
	char value[1024]; /* should enough */
	int valuesize = sizeof(value);
	
	return mydbm_fetch(db, key, value, &valuesize) + 1;
		/* -2: return -1; case -1: return 0; case 0: return 1; */
#else
	datum akey;
	
	akey.dsize = strlen(akey.dptr = key);
	return gdbm_exists((*db), akey) ? 1 : 0;
#endif
}

int mydbm_delete(MYDBM *db, char *key) {
	datum akey;
	int result;
	
	akey.dsize = strlen(akey.dptr = key);
#ifdef GDBM
	result = gdbm_delete((*db), akey);
#else
	result = dbm_delete((*db), akey);
#endif
	if (result < 0) {
#ifdef MYDBM_SPDBM
		fprintf(stderr, "mydbm: can't delete key/value from "
			"database\n");
#else
		pop_log(pop_priority, "mydbm: can't delete key/value from "
			"database");
#endif
		return -1;
	}
	return 0;
}

int mydbm_store(MYDBM *db, char *key, char *value, int valuesize, int replace) {
	datum akey, avalue;
	int result;
	
	akey.dsize = strlen(akey.dptr = key);
	avalue.dptr = value;
	avalue.dsize = valuesize;
#ifdef GDBM
	result = gdbm_store((*db), akey, avalue, (replace ? GDBM_REPLACE :
		GDBM_INSERT));
#else
	result = dbm_store((*db), akey, avalue, (replace ? DBM_REPLACE :
		DBM_INSERT));
#endif
	if (result < 0) {
#ifdef MYDBM_SPDBM
		fprintf(stderr, "mydbm: can't store key/value to database\n");
#else
		pop_log(pop_priority, "mydbm: can't store key/value to "
			"database");
#endif
		return -1;
	}
	return result;
}

int mydbm_entry(MYDBM *db, char *key, int keylength, char *value,
	int *valuelength, int first) {
	datum akey;
#ifdef GDBM
	static datum lastkey;

	if (first) {
		lastkey.dptr = NULL;
		lastkey.dsize = 0;
	}
	if (first)
		akey = gdbm_firstkey(*db);
	else
		akey = gdbm_nextkey(*db, lastkey);
	if (lastkey.dptr != NULL)
		free(lastkey.dptr);
#else
	if (first)
		akey = dbm_firstkey(*db);
	else
		akey = dbm_nextkey(*db);
#endif
	if ((akey.dsize >= keylength) && (akey.dptr != NULL)) {
#ifdef MYDBM_SPDBM
		fprintf(stderr, "key is too long - database damaged\n");
#else
		pop_log(pop_priority, "key is too long - database damaged");
#endif
		return -1;
	}
	if (akey.dptr != NULL) {
#ifdef GDBM
		lastkey = akey;
#endif
		memcpy(key, akey.dptr, akey.dsize);
		key[akey.dsize] = 0;
	} else
		return 0; /* no entries */
	if (mydbm_fetch(db, key, value, valuelength) < 0) {
#ifdef GDBM
		free(lastkey.dptr);
#endif
		return -1; /* no entries */
	} else
		return 1;
}

int mydbm_nextentry(MYDBM *db, char *key, int keylength, char *value,
		    int *valuelength) {
	return mydbm_entry(db, key, keylength, value, valuelength, 0);
}

int mydbm_firstentry(MYDBM *db, char *key, int keylength, char *value,
		    int *valuelength) {
	return mydbm_entry(db, key, keylength, value, valuelength, 1);
}
