/* $Id: sktbuff.c,v 1.11 2000/09/29 10:31:54 jurekb Exp $ */
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <fcntl.h>
#include <errno.h>

#include "sktbuff.h"
#include "fdfgets.h"
#include "log.h"

#ifdef SPENCRYPTION

# ifdef SPSASL
#  include "spsasl.h"
#  include <sasl.h>
#  include <saslutil.h>
extern sasl_conn_t *sasl_connection;
char *dec_output = NULL, *dec_act = NULL;
unsigned int dec_length = 0;
int first_response = 0; /* message that shouldn't be encrypted by SASL */
# endif

# ifdef SPTLS
#  include "sptls.h"
# endif

int enctrans = 0, encenabled = 0; /* write through pipe */
int encpipe[2] = {-1, -1}, decpipe[2] = {-1, -1}; /* pipe descriptors */
#endif /* SPENCRYPTION */

char inbuf[512];
size_t inbuf_count = 0;
char *inbuf_act = inbuf;

char outbuf[8192]; /* to reduce system calls, it may be > PIPE_BUF */
size_t outbuf_count = 0;
ssize_t outbuf_crlf = -1;
int outbuf_err = 0;

#ifdef SPENCRYPTION
int open_encpipes(void) {
	if (pipe(encpipe) < 0) {
		pop_log(pop_priority, "sktbuff: can't create pipe");
		pop_error("sktbuff: pipe");
		return -1;
	}
	if (pipe(decpipe) < 0) {
		pop_log(pop_priority, "sktbuff: can't create pipe");
		pop_error("sktbuff: pipe");
		close(encpipe[0]);
		close(encpipe[1]);
		return -1;
	}
	return 0;
}

void close_encpipes(void) {
	if (encpipe[0] >= 0)
		close(encpipe[0]);
	if (encpipe[1] >= 0)
		close(encpipe[1]);
	if (decpipe[0] >= 0)
		close(decpipe[0]);
	if (decpipe[1] >= 0)
		close(decpipe[1]);
	encpipe[0] = encpipe[1] = decpipe[0] = decpipe[1] = -1;
}

void server_encpipes(void) { /* encryption server reads data from encpipe
				and write to decpipe */
	if (encpipe[1] >= 0)
		close(encpipe[1]);
	if (decpipe[0] >= 0)
		close(decpipe[0]);
	encpipe[1] = decpipe[0] = -1;
}

void client_encpipes(void) {
	if (encpipe[0] >= 0)
		close(encpipe[0]);
	if (decpipe[1] >= 0)
		close(decpipe[1]);
	encpipe[0] = decpipe[1] = -1;
}

ssize_t skt_read(int fd, void *buf, size_t size) { /* fd is ignored */
	if (enctrans)
		return read(decpipe[0], buf, size);
	if (encenabled)
		return enc_read(buf, size, 1); /* block until data arrives */
	return read(0, buf, size);
}


void encryption_server(void) {
	fd_set rdset;
	char encbuf[OUTBUFSIZE];
	char decbuf[OUTBUFSIZE]; /* OUTBUFSIZE <= PIPE_BUF */
	size_t pending = 0;
	ssize_t result;
	int es_eof = 0;
	
	if (fcntl(decpipe[1], F_SETFL, O_NONBLOCK) < 0) {
		pop_error("sktbuff: fcntl");
		return;
	}
	if (!encenabled)
		return;
	while(1) {
		FD_ZERO(&rdset);
		if ((pending == 0) && (!es_eof))
			/* if decpipe is full we must wait until server flushes
			   his output through encpipe */
			FD_SET(0, &rdset);
		FD_SET(encpipe[0], &rdset);
		if (select(encpipe[0] + 1, &rdset, NULL, NULL, NULL) < 0) {
			memset(decbuf, 0, sizeof(decbuf));
			pop_log(pop_priority, "sktbuff: select failed");
			pop_error("sktbuff: select");
			return;
		}
		if (FD_ISSET(encpipe[0], &rdset)) { /* flush server responses */
			if ((result = read(encpipe[0], encbuf, sizeof(encbuf)))
				<= 0) {
				if (result < 0) {
					memset(decbuf, 0, sizeof(decbuf));
					pop_log(pop_priority, "sktbuff: can't "
						"read from pipe");
					pop_error("sktbuff: read");
					return;
				}
				/* encpipe eof */
				return;
			}
			if (enc_write(encbuf, result) < 0) {
				memset(decbuf, 0, sizeof(decbuf));
				pop_log(pop_priority, "sktbuff: can't write "
					"to socket");
				pop_error("sktbuff: write");
				return;
			}
		}
		if (FD_ISSET(0, &rdset) || dec_pending() || (pending > 0)) {
			if (pending > 0) {
				result = write(decpipe[1], decbuf, pending);
				if ((result < 0) && (errno == EAGAIN))
					continue;
				memset(decbuf, 0, sizeof(decbuf));
				if (result < 0) {
					pop_log(pop_priority, "sktbuff: can't "
						"write to socket");
					pop_error("sktbuff: write");
					return;
				}
				pending = 0;
			}
			if ((!FD_ISSET(0, &rdset)) && (!dec_pending()))
				continue; /* no data, pending was > 0  */
			do {
				if ((result = enc_read(decbuf, sizeof(decbuf),
					0)) == -1) { /* error */
					memset(decbuf, 0, sizeof(decbuf));
					pop_log(pop_priority, "sktbuff: can't "
						"read from socket");
					pop_error("sktbuff: read");
					return;
				}
				if (result == 0) { /* socket eof */
					es_eof = 1;
					continue;
				}
				if (result > 0)
					if (write(decpipe[1], decbuf, result) <
						0) {
						if (errno == EAGAIN) {
							pending = result;
							continue;
						}
						memset(decbuf, 0, sizeof(decbuf));
						pop_log(pop_priority,
							"sktbuff: can't write "
							"to pipe");
						pop_error("sktbuff: write");
						return;
					}
				/* if result == -2 continue */
			} while (dec_pending());
		}
	}
	memset(decbuf, 0, sizeof(decbuf));
}




ssize_t enc_write(void *buf, size_t size) {
/* encenabled != 0, ((encenabled & 1) || (encenabled & 2)) */

	ssize_t result;
	char *enc_output;
	unsigned int enc_length;
	
# ifdef SPTLS
#  ifdef SPSASL
	if (((encenabled & 2) && first_response) ||
	    (encenabled == 2)) {
		first_response = 0;
		return tls_write(buf, size);
	}
#  else
	/* TLS without SASL and (encenabled & 2) */
	return tls_write(buf, size);
#  endif
# endif

# ifdef SPSASL
	if (first_response) { /* we know TLS is not active */
		first_response = 0;
		return write(1, buf, size);
	} /* (encenabled & 1) */
	if ((result = sasl_encode(sasl_connection, (char *) buf, size,
				  &enc_output, &enc_length)) != SASL_OK) {
		pop_log(pop_priority, "sktbuff: %.512s", sasl_errstring(result,
			NULL, NULL));
		return -1;
	}
	result = 0;
#  ifdef SPTLS
	if (enc_length > 0) {
		if (encenabled & 2)
			result = tls_write(enc_output, enc_length);
		else
			result = write(1, enc_output, enc_length);
	}
#  else /* SPTLS */
	if (enc_length > 0)
		result = write(1, enc_output, enc_length);
#  endif /* SPTLS */
	if (enc_output != NULL)
		free(enc_output);
	return ((result <= 0) ? result : size);
# endif /* SPSASL */
}


#ifdef SPSASL
void enc_cleanup(void) {
	if (dec_output == NULL)
		return;
	memset(dec_act, 0, dec_length);
	free(dec_output);
}
#endif

ssize_t enc_read(void *buf, size_t size, int block) {
/* return values:
	0 - end of file
	-1 - error
	-2 - there are data in socket, but not enough to decode some bytes
	-2 is returned only when block == 0
*/
	ssize_t result, readsize;
	char readbuf[2048];

# ifdef SPTLS
	if (encenabled == 2)
		return tls_read(buf, size, block);
# endif /* SPTLS */

# ifdef SPSASL
	if (dec_length > 0) { /* SASL encryption is used */
		result = ((dec_length < size) ? dec_length : size);
		memcpy(buf, dec_act, result);
		memset(dec_act, 0, result);
		dec_act += result;
		dec_length -= result;
		if (dec_length == 0) {
			free(dec_output);
			dec_output = dec_act = NULL;
		}
		return result;
	}
	while (dec_length == 0) {
#  ifdef SPTLS
		if (encenabled & 2) {
			if ((readsize = tls_read(readbuf, sizeof(readbuf),
				block)) <= 0)
				return readsize;
		} else
			if ((readsize = read(0, readbuf, sizeof(readbuf)))
				<= 0)
				return readsize;
#  else
		if ((readsize = read(0, readbuf, sizeof(readbuf))) <= 0)
			return readsize;
#  endif
		if ((result = sasl_decode(sasl_connection, readbuf, readsize,
			&dec_output, &dec_length)) != SASL_OK) {
			memset(readbuf, 0, sizeof(readbuf));
			pop_log(pop_priority, "sktbuff: %.512s",
				sasl_errstring(result, NULL, NULL));
			return -1;
		}
		memset(readbuf, 0, sizeof(readbuf));
		if ((dec_length == 0) && (dec_output != NULL)) {
			free(dec_output);
			dec_output = NULL;
		}
		if ((dec_length == 0) && (block == 0)) /* && readsize > 0 */
			return -2;
	}
	/* dec_length > 0, so ... */
	dec_act = dec_output;
	return enc_read(buf, size, block); /* dec_length > 0 */
# endif /* SPSASL */
}

int dec_pending(void) {
	int apending = 0;

# ifdef SPSASL
	if (encenabled & 1)
		apending = ((dec_length == 0) ? 0 : 1);
# endif
# ifdef SPTLS
	if ((!apending) && (encenabled & 2))
		apending = tls_pending();
# endif
	return apending;
}

#endif /* SPENCRYPTION */

ssize_t skt_fgets(char *s, size_t size) {
	return gen_fgets(s, size, 0, skt_read, inbuf, &inbuf_count, &inbuf_act,
		sizeof(inbuf), 1, NULL);
}

void skt_cleanup(void) {
	memset(inbuf, 0, sizeof(inbuf));
}

int buf_flush(void) { /* called in spop3d_done(), atexit(spop3d_done),
			 at end of greeting and every command */
	size_t count = outbuf_count;

	if ((outbuf_err != 0) || (outbuf_count == 0))
		return 0;
	outbuf_count = 0;
	outbuf_crlf = -1;
#ifdef SPENCRYPTION
	if (enctrans) {
		if (write(encpipe[1], outbuf, count) < 0) {
			outbuf_err = 1;
			pop_error("sktbuff: write");
			return -1;
		}
	} else {
		if (encenabled) {
			if (enc_write(outbuf, count) < 0) {
				outbuf_err = 1;
				pop_error("sktbuff: write");
				return -1;
			}
		} else
			if (write(1, outbuf, count) < 0) {
				outbuf_err = 1;
				pop_error("sktbuff: write");
				return -1;
			}
	}
#else
	if (write(1, outbuf, count) < 0) {
		outbuf_err = 1;
		pop_error("sktbuff: write");
		return -1;
	}
#endif
	return 0;
}

ssize_t skt_write(void *buf, size_t size) {
	char *abuf = (char *) buf;
	size_t count;
	ssize_t cwrite = 0, result;

	if (outbuf_err != 0)
		return -1;
	while (size > 0) {
		count = (((sizeof(outbuf) - outbuf_count) < size) ?
			 (sizeof(outbuf) - outbuf_count) : size);
		memcpy(outbuf + outbuf_count, abuf, count);
		size -= count;
		outbuf_count += count;
		abuf += count;
		cwrite += count;
		if (outbuf_count == sizeof(outbuf)) {
#ifdef SPENCRYPTION
			if (enctrans)
				result = write(encpipe[1], outbuf,
					sizeof(outbuf));
			else {
				if (encenabled)
					result = enc_write(outbuf,
						sizeof(outbuf));
				else
					result = write(1, outbuf,
						sizeof(outbuf));
			}
#else
			result = write(1, outbuf, sizeof(outbuf));
#endif
			outbuf_count = 0;
			if (result < 0) {
				outbuf_err = 1;
				return -1;
			}
		}
	}
	return cwrite;
}
